/*
 * Copyright (C) 2001 USAGI/WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <netinet/in.h>
#include <netdb.h>
#include <getopt.h>

int main(int argc, char **argv){
	const char *myname = argv[0];
	struct msghdr msghdr;
	struct iovec iov[2];
	char buf[65536];
	struct sockaddr_storage sa;
	int salen;
	int s, cc;
	int c;
	int v6only, setv6only = 0;
	struct addrinfo hints, *res0;
	int gai;
	char *host, *serv;
	char abuf[NI_MAXHOST];

	while((c = getopt(argc, argv, "6:")) != EOF) {
		switch(c) {
		case '6':
			setv6only = 1;
			v6only = atoi(optarg);
			break;
		}
	}

	argc -= optind;
	argv += optind;

	if (argc == 0) {
		printf("%s [-6 v6only-sockopt] [host] port\n", myname);
		exit(1);
	}

        host = argc > 1 ? argv[0] : NULL;
        serv = argc > 1 ? argv[1] : argv[0];

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = PF_UNSPEC;
	hints.ai_socktype = SOCK_DGRAM;
	hints.ai_flags = AI_PASSIVE;
	gai = getaddrinfo(host, serv, &hints, &res0);
	if (gai) {
		printf("getaddrinfo(): %s\n",
			gai_strerror(gai));
		exit(2);
	}

	s = socket(res0->ai_family, res0->ai_socktype, res0->ai_protocol);
	if (s < 0) {
		perror("socket");
		exit(2);
	}

	if (setv6only) {
		printf("set IPV6_V6ONLY to %d\n", v6only);
#ifdef IPV6_V6ONLY
		if (setsockopt(s, IPPROTO_IPV6, IPV6_V6ONLY, &v6only, sizeof(v6only)) < 0)
			perror("setsockopt(IPV6_V6ONLY)");
#else
		fprintf(stderr, "IPV6_V6ONLY not supported.\n");
#endif
	}

	if (bind(s, res0->ai_addr, res0->ai_addrlen)) {
		freeaddrinfo(res0);
		perror("bind");
		exit(2);
	}
	freeaddrinfo(res0);

	memset(&iov, 0, sizeof(iov));
	iov[0].iov_base = buf;
	iov[0].iov_len = sizeof(buf);

	memset(&msghdr, 0, sizeof(msghdr));
	msghdr.msg_name = (void *)&sa;
	msghdr.msg_namelen = sizeof(sa);
	msghdr.msg_iov = iov;
	msghdr.msg_iovlen = 1;
	msghdr.msg_control = NULL;
	msghdr.msg_controllen = 0;
	msghdr.msg_flags = 0;

	cc = recvmsg(s, &msghdr, 0);
	if (cc < 0) {
		perror("recvmsg");
		exit(2);
	}
	salen = msghdr.msg_namelen;

	if (getnameinfo((struct sockaddr *)&sa, salen,
			abuf, sizeof(abuf), NULL, 0,
			NI_NUMERICHOST))
		strcpy(abuf, "???");

	printf("received %d byte(s) from %s\n", cc, abuf);

	exit(0);
}

